#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "transform.h"
#include "camera.h"
#include "primitives.h"

#define sqr(x) ((x)*(x))
#define EPS 0.0000001
#define INFTY 10e9

char infilename[1000];
int VERBOSE=0,DEBUG=0;
int XSIZE=640,YSIZE=480;
int POCET_ITERACII=2000;
double MAX_TRIANGLE=5000.0;
int DO_MONTE_CARLO=1;
int MONTE_CARLO_RAYS=4;
FILE *f,*fvis;
int READ_VIS=0;
int SMOOTH=0;

tCamera Kamera;
tTriangle TriangleList[3000]; int TriangleCount=0;
double Tarea[3000];
double FF[3000][3000];
int iteracia;

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   POMOCNE HALUZE                                          //
///////////////////////////////////////////////////////////////////////////////////////////////

// ak sme verbose, vypise message 
void message(const char *format, ...){ va_list ap; va_start(ap,format); if (VERBOSE) vprintf(format,ap); va_end(ap); }
void dbgmessage(const char *format, ...){ va_list ap; va_start(ap,format); if (DEBUG) vprintf(format,ap); va_end(ap); }

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   INTERPOLACIA NORMAL                                     //
///////////////////////////////////////////////////////////////////////////////////////////////

void InterpolateAnyTriangleNormal(tTriangle T, tPoint bod, tVector *normal){
  double u,v;
  tVector smer1,smer2,vek;
  int i;
  for (i=0;i<3;i++) {
    vek[i]=bod[i]-T.vertex[0][i];
    smer1[i]=T.vertex[1][i]-T.vertex[0][i];
    smer2[i]=T.vertex[2][i]-T.vertex[0][i];
  }
  if (fabs(smer1[0])+fabs(smer2[0]) < EPS) { smer1[0]=smer1[2]; smer2[0]=smer2[2]; }
  if (fabs(smer1[0]*smer2[1] - smer1[1]*smer2[0])<EPS) { smer1[1]=smer1[2]; smer2[1]=smer2[2]; }
  u=(vek[0]*smer2[1] - vek[1]*smer2[0]) / (smer1[0]*smer2[1] - smer1[1]*smer2[0]);
  v=(vek[0]*smer1[1] - vek[1]*smer1[0]) / (smer2[0]*smer1[1] - smer2[1]*smer1[0]);
//  dbgmessage("interpolation: u=%.3f v=%.3f\n",u,v);
  for (i=0;i<3;i++) 
    (*normal)[i]=u*T.normal[1][i]+v*T.normal[2][i]+(1-u-v)*T.normal[0][i];
  
  Normalize(normal);
}

void InterpolateTriangleNormal(int tr, tPoint bod, tVector *normal){
  InterpolateAnyTriangleNormal(TriangleList[tr],bod,normal);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   NACITANIE                                               //
///////////////////////////////////////////////////////////////////////////////////////////////

void LoadCamera(void){
  char row[1000],*rest;
  int i,rl;
  tVector smer,perp;
  double pom;
 
  Kamera.location[0]=Kamera.location[1]=Kamera.location[2]=0; Kamera.location[3]=1;
  Kamera.direction[2]=1; Kamera.direction[0]=Kamera.direction[1]=Kamera.direction[3]=0;
  Kamera.right[0]=0.3*1.33333333; Kamera.right[1]=Kamera.right[2]=Kamera.right[3]=0;
  Kamera.up[1]=0.3*1; Kamera.up[0]=Kamera.up[2]=Kamera.up[3]=0;
  
  fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  while (strcmp(row,"end")) {
    while (1) { // trick, bude len raz, ale chceme vediet breakovat
      if ((row[0]=='/')&&(row[1]=='/')) break;
      rl=strlen(row); for (i=0;i<rl;i++) if (row[i]=='=') break;
      if (i==rl) { dbgmessage("Load: camera: I don't understand: '%s'\n",row); break; }
      rest=row+i+1; row[i]=0;
      if (!strcmp(row,"location")) {
        sscanf(rest,"%lf , %lf , %lf",&Kamera.location[0],&Kamera.location[1],&Kamera.location[2]);
        dbgmessage("Load: camera: location %.2f %.2f %.2f\n",Kamera.location[0],Kamera.location[1],Kamera.location[2]);
        break;
      }
      if (!strcmp(row,"look_at")) {
        sscanf(rest,"%lf , %lf , %lf",&smer[0],&smer[1],&smer[2]); smer[3]=0;
        for (i=0;i<3;i++) { smer[i]-=Kamera.location[i]; Kamera.direction[i]=perp[i]=smer[i]; }
        perp[1]=0;
        if (Length(perp)<EPS){
          // pozerame hore alebo dole
        } else {
          pom=perp[0]; perp[0]=perp[2]; perp[2]=-pom; // note the minus sign!
          CrossProduct(smer,perp,&Kamera.up); Normalize(&Kamera.up);
          CrossProduct(Kamera.up,Kamera.direction,&Kamera.right);
          for (i=0;i<3;i++) { 
            Kamera.up[i]*=0.3*Length(Kamera.direction);
            Kamera.right[i]*=0.4*Length(Kamera.direction);
          }
        }
        dbgmessage("Load: camera: direction %.2f %.2f %.2f\n",Kamera.direction[0],Kamera.direction[1],Kamera.direction[2]);
        dbgmessage("Load: camera: up %.2f %.2f %.2f\n",Kamera.up[0],Kamera.up[1],Kamera.up[2]);
        dbgmessage("Load: camera: right %.2f %.2f %.2f\n",Kamera.right[0],Kamera.right[1],Kamera.right[2]);
        break;
      }
      if (!strcmp(row,"up")) {
        sscanf(rest,"%lf , %lf , %lf",&Kamera.up[0],&Kamera.up[1],&Kamera.up[2]);
          CrossProduct(Kamera.up,Kamera.direction,&Kamera.right);
          Normalize(&Kamera.right); 
          for (i=0;i<3;i++) Kamera.right[i]*=4/3.0*Length(Kamera.up);
        break;
      }
      dbgmessage("Load: camera: Not implemented: '%s'\n",row);
      break; 
    }
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  }
}

void AddTriangle(tTriangle T){
  tVector V1,V2,V;
  tTriangle pom;
  int i;
  
  for (i=0;i<3;i++) V1[i]=T.vertex[1][i]-T.vertex[0][i];
  for (i=0;i<3;i++) V2[i]=T.vertex[2][i]-T.vertex[0][i];
  CrossProduct(V1,V2,&V);
  message("V = %g\n",Length(V));
  if (Length(V)*0.5<=MAX_TRIANGLE) {
    // add
    Tarea[TriangleCount]=Length(V)*0.5;
    TriangleList[TriangleCount++]=T;
  } else {
    // subdivide
    pom=T;
    for (i=0;i<3;i++) pom.vertex[1][i]=(T.vertex[0][i]+T.vertex[1][i])/2.0; 
    for (i=0;i<3;i++) pom.vertex[2][i]=(T.vertex[0][i]+T.vertex[2][i])/2.0;
    InterpolateAnyTriangleNormal(T,pom.vertex[1],&pom.normal[1]);
    InterpolateAnyTriangleNormal(T,pom.vertex[2],&pom.normal[2]);
    AddTriangle(pom);

    pom=T;
    for (i=0;i<3;i++) pom.vertex[0][i]=(T.vertex[0][i]+T.vertex[1][i])/2.0; 
    for (i=0;i<3;i++) pom.vertex[2][i]=(T.vertex[1][i]+T.vertex[2][i])/2.0;
    InterpolateAnyTriangleNormal(T,pom.vertex[0],&pom.normal[0]);
    InterpolateAnyTriangleNormal(T,pom.vertex[2],&pom.normal[2]);
    AddTriangle(pom);

    pom=T;
    for (i=0;i<3;i++) pom.vertex[0][i]=(T.vertex[0][i]+T.vertex[2][i])/2.0; 
    for (i=0;i<3;i++) pom.vertex[1][i]=(T.vertex[1][i]+T.vertex[2][i])/2.0;
    InterpolateAnyTriangleNormal(T,pom.vertex[0],&pom.normal[0]);
    InterpolateAnyTriangleNormal(T,pom.vertex[1],&pom.normal[1]);
    AddTriangle(pom);
    
    pom=T;
    for (i=0;i<3;i++) pom.vertex[0][i]=(T.vertex[0][i]+T.vertex[1][i])/2.0; 
    for (i=0;i<3;i++) pom.vertex[1][i]=(T.vertex[1][i]+T.vertex[2][i])/2.0;
    for (i=0;i<3;i++) pom.vertex[2][i]=(T.vertex[2][i]+T.vertex[0][i])/2.0;
    InterpolateAnyTriangleNormal(T,pom.vertex[0],&pom.normal[0]);
    InterpolateAnyTriangleNormal(T,pom.vertex[1],&pom.normal[1]);
    InterpolateAnyTriangleNormal(T,pom.vertex[2],&pom.normal[2]);
    AddTriangle(pom);
  }
}

void LoadTriangle(void){
  char row[1000],*rest;
  int i,rl,v,ma_normaly=0;
  tVector v1,v2;
  tTriangle pomTriangle;
 
  // NASTAVIME DEFAULT HODNOTY FARBY A MATERIALU
  pomTriangle.diffuseColor[0]=pomTriangle.diffuseColor[1]
    =pomTriangle.diffuseColor[2]=1.0;
  pomTriangle.emission[0]=pomTriangle.emission[1]
    =pomTriangle.emission[2]=0.0;
  pomTriangle.incidentLight[0]=pomTriangle.incidentLight[1]
    =pomTriangle.incidentLight[2]=0.0;
  pomTriangle.excidentLight[0]=pomTriangle.excidentLight[1]
    =pomTriangle.excidentLight[2]=0.0;
  pomTriangle.id=-1;
  
  fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  while (strcmp(row,"end")) {
    while (1) { // trick, bude len raz, ale chceme vediet breakovat
      if ((row[0]=='/')&&(row[1]=='/')) break;
      rl=strlen(row); for (i=0;i<rl;i++) if (row[i]=='=') break;
      if (i==rl) { dbgmessage("Load: triangle: I don't understand: '%s'\n",row); break; }
      rest=row+i+1; row[i]=0;
      if (sscanf(row,"vertex%d",&v)) {
        sscanf(rest,"%lf , %lf , %lf",&pomTriangle.vertex[v][0],
            &pomTriangle.vertex[v][1],&pomTriangle.vertex[v][2]);
        dbgmessage("Load: triangle: vertex %.2f %.2f %.2f\n",pomTriangle.vertex[v][0],
            pomTriangle.vertex[v][1],pomTriangle.vertex[v][2]);
        break;
      }
      if (sscanf(row,"normal%d",&v)) {
        sscanf(rest,"%lf , %lf , %lf",&pomTriangle.normal[v][0],
            &pomTriangle.normal[v][1],&pomTriangle.normal[v][2]);
        dbgmessage("Load: triangle: normal %.2f %.2f %.2f\n",pomTriangle.normal[v][0],
            pomTriangle.normal[v][1],pomTriangle.normal[v][2]);
        ma_normaly=1;
        break;
      }
      if (!strcmp(row,"color")) {
        sscanf(rest,"%lf , %lf , %lf",&pomTriangle.diffuseColor[0],
            &pomTriangle.diffuseColor[1],&pomTriangle.diffuseColor[2]);
        break;
      }
      if (!strcmp(row,"id")) {
        sscanf(rest,"%d",&pomTriangle.id);
        break;
      }
      if (!strcmp(row,"emission")) {
        sscanf(rest,"%lf , %lf , %lf",&pomTriangle.emission[0],
            &pomTriangle.emission[1],&pomTriangle.emission[2]);
        break;
      }
      dbgmessage("Load: triangle: Not implemented: '%s'\n",row);
      break; 
    }
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  }
  if (!ma_normaly){
    for (i=0;i<3;i++) { 
      v1[i]=pomTriangle.vertex[1][i]-pomTriangle.vertex[0][i];
      v2[i]=pomTriangle.vertex[2][i]-pomTriangle.vertex[0][i];
    }
    for (i=0;i<3;i++) {
      CrossProduct(v2,v1,&pomTriangle.normal[i]);
      Normalize(&pomTriangle.normal[i]);
      dbgmessage("New normal: %.3f %.3f %.3f\n",
          pomTriangle.normal[i][0],
          pomTriangle.normal[i][1],
          pomTriangle.normal[i][2]);
    }
  }
  for (i=0;i<3;i++) pomTriangle.excidentLight[i]=pomTriangle.emission[i];
  AddTriangle(pomTriangle);
}

void LoadScene(void){
  int rl;
  char row[1000];
  
  if (!(f=fopen(infilename,"r"))) { message("Open failed.\n"); exit(1); }
  while (!feof(f)){
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
    dbgmessage("Debug: loaded '%s'\n",row);
    if (strlen(row)==0) continue;
    if ((row[0]=='/')&&(row[1]=='/')) continue;
    if (!strcmp(row,"camera")) { LoadCamera(); continue; }
    if (!strcmp(row,"triangle")) { LoadTriangle(); continue; }
    if (!strcmp(row,"eof")) break; 
    dbgmessage("Load: I don't understand: '%s'\n",row);
  }
  fclose(f);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   PRIESECNIKY                                             //
///////////////////////////////////////////////////////////////////////////////////////////////

void IntersectTriangle(int trojuh, tPoint source, tVector smer, double *where){
  tVector edge1,edge2,pvec,tvec,qvec,ssmer;
  int i;
  double det,inv,u,v,l;

  for (i=0;i<3;i++) { 
    edge1[i]=TriangleList[trojuh].vertex[1][i]-TriangleList[trojuh].vertex[0][i]; 
    edge2[i]=TriangleList[trojuh].vertex[2][i]-TriangleList[trojuh].vertex[0][i]; 
    ssmer[i]=smer[i];
  }
  l=Length(ssmer); Normalize(&ssmer);
  CrossProduct(ssmer,edge2,&pvec);
  det=DotProduct(edge1,pvec);
  if (fabs(det)<EPS) { (*where)=INFTY; return; } // sme v rovine trojuholnika
  inv=1/det;
  for (i=0;i<3;i++) tvec[i]=source[i]-TriangleList[trojuh].vertex[0][i]; 
  u=DotProduct(tvec,pvec)*inv;
  if ((u<EPS)||(u>1-EPS)) { (*where)=INFTY; return; } // sur. u je mimo
  CrossProduct(tvec,edge1,&qvec);
  v=DotProduct(ssmer,qvec)*inv;
  if ((v<EPS)||(u+v>1-EPS)) { (*where)=INFTY; return; } // sur. v je mimo
  (*where)=DotProduct(edge2,qvec)*inv/l;
}

void IntersectTriangle2(int trojuh, tPoint source, tVector smer, double *where, double *u, double *v){
  tVector edge1,edge2,pvec,tvec,qvec,ssmer;
  int i;
  double det,inv,l;

  for (i=0;i<3;i++) { 
    edge1[i]=TriangleList[trojuh].vertex[1][i]-TriangleList[trojuh].vertex[0][i]; 
    edge2[i]=TriangleList[trojuh].vertex[2][i]-TriangleList[trojuh].vertex[0][i]; 
    ssmer[i]=smer[i];
  }
  l=Length(ssmer); Normalize(&ssmer);
  CrossProduct(ssmer,edge2,&pvec);
  det=DotProduct(edge1,pvec);
  if (fabs(det)<EPS) { (*where)=INFTY; return; } // sme v rovine trojuholnika
  inv=1/det;
  for (i=0;i<3;i++) tvec[i]=source[i]-TriangleList[trojuh].vertex[0][i]; 
  (*u)=DotProduct(tvec,pvec)*inv;
  if (((*u)<EPS)||((*u)>1-EPS)) { (*where)=INFTY; return; } // sur. u je mimo
  CrossProduct(tvec,edge1,&qvec);
  (*v)=DotProduct(ssmer,qvec)*inv;
  if (((*v)<EPS)||((*u)+(*v)>1-EPS)) { (*where)=INFTY; return; } // sur. v je mimo
  (*where)=DotProduct(edge2,qvec)*inv/l;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   RAYTRACING                                              //
///////////////////////////////////////////////////////////////////////////////////////////////

void CastRay(tPoint source, tVector smer, tVector* farba, int vnorenie){
  double nearest,where,u,v,uu,vv;
  int i,co;
  
  vnorenie=0; // aby nehadzal warning ;)

  if (READ_VIS) { 
    fscanf(fvis,"%d ",&co); 
    if (co==-1) {
      (*farba)[0]=(*farba)[1]=(*farba)[2]=0.0;
    } else {
      IntersectTriangle2(co,source,smer,&where,&u,&v);  
/*      message("u=%g v=%g\n",u,v);
      message("  vertexcolor0= %g %g %g\n",
          TriangleList[co].vertexColor[0][0],
          TriangleList[co].vertexColor[0][1],
          TriangleList[co].vertexColor[0][2]);
      message("  vertexcolor1= %g %g %g\n",
          TriangleList[co].vertexColor[1][0],
          TriangleList[co].vertexColor[1][1],
          TriangleList[co].vertexColor[1][2]);
      message("  vertexcolor2= %g %g %g\n",
          TriangleList[co].vertexColor[2][0],
          TriangleList[co].vertexColor[2][1],
          TriangleList[co].vertexColor[2][2]);*/
      // nasli sme priesecnik
      (*farba)[0]=(*farba)[1]=(*farba)[2]=0.0;
      for (i=0;i<3;i++) (*farba)[i]+=      u*TriangleList[co].vertexColor[1][i];
      for (i=0;i<3;i++) (*farba)[i]+=      v*TriangleList[co].vertexColor[2][i];
      for (i=0;i<3;i++) (*farba)[i]+=(1-u-v)*TriangleList[co].vertexColor[0][i];
      // upravime
      for (i=0;i<3;i++) if ((*farba)[i]<0.0) (*farba)[i]=0.0; 
      for (i=0;i<3;i++) if ((*farba)[i]>1.0) (*farba)[i]=1.0; 
//      message("  farba= %g %g %g\n",(*farba)[0],(*farba)[1],(*farba)[2]);
    }
  } else {
    nearest=INFTY;
    for (i=0;i<TriangleCount;i++) { 
      IntersectTriangle2(i,source,smer,&where,&uu,&vv);  
      if ((where>EPS) && (where<nearest)) { nearest=where; co=i; u=uu; v=vv; } 
    }
  
    if (INFTY-nearest>EPS) {
      // nasli sme priesecnik
      (*farba)[0]=(*farba)[1]=(*farba)[2]=0.0;
      for (i=0;i<3;i++) (*farba)[i]+=      u*TriangleList[co].vertexColor[1][i];
      for (i=0;i<3;i++) (*farba)[i]+=      v*TriangleList[co].vertexColor[2][i];
      for (i=0;i<3;i++) (*farba)[i]+=(1-u-v)*TriangleList[co].vertexColor[0][i];
      // upravime
      for (i=0;i<3;i++) if ((*farba)[i]<0.0) (*farba)[i]=0.0; 
      for (i=0;i<3;i++) if ((*farba)[i]>1.0) (*farba)[i]=1.0; 
      fprintf(fvis,"%d ",co);
    } else {
      (*farba)[0]=(*farba)[1]=(*farba)[2]=0.0;
      fprintf(fvis,"%d ",-1);
    }

  }
}

void RayTrace(void){
  FILE *fout;
  tVector farba,smer;
  int i,j,k,l;
  double x,y;
  char fn[1000];

  fout=fopen("output.bmp","w");
  fprintf(fout,"%c%c%c%c",0x42,0x4D,0x36,0x10);
  fprintf(fout,"%c%c%c%c",0x0E,0x00,0x00,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x36,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x28,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,XSIZE%256,XSIZE/256);
  
  fprintf(fout,"%c%c%c%c",0x00,0x00,YSIZE%256,YSIZE/256);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x01,0x00);
  fprintf(fout,"%c%c%c%c",0x18,0x00,0x00,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x00,0x10);
  fprintf(fout,"%c%c%c%c",0x0E,0x00,0x6D,0x0B);
  
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x6D,0x0B);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x00,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x00,0x00);
  fprintf(fout,"%c%c",0x00,0x00);
  
  sprintf(fn,"%s-%.5f-visibility.dat",infilename,MAX_TRIANGLE);
  fvis=fopen(fn,"r"); if (fvis==NULL) { fvis=fopen(fn,"w"); READ_VIS=0; } else { READ_VIS=1; }

  for (j=0;j<YSIZE;j++) { 
    for (i=0;i<XSIZE;i++) {
      // spocitame si bod sveta, cez ktory ideme 
      x=2*i-XSIZE; y=2*j-YSIZE; x/=XSIZE; y/=YSIZE;
      for (k=0;k<3;k++) smer[k]=Kamera.direction[k]+x*Kamera.right[k]+y*Kamera.up[k];
      //message("smer: %.2f %.2f %.2f\n",smer[0],smer[1],smer[2]);
      CastRay(Kamera.location,smer,&farba,1);
      for (l=2;l>=0;l--) { k=(int)(farba[l]*255); putc((char)k,fout); }
    }
    message("Rows done: %d\r",j+1); fflush(stdout);
  }
  fclose(fout); fclose(fvis);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   FORM FAKTORY                                            //
///////////////////////////////////////////////////////////////////////////////////////////////

int LoadFormFactors(void){
  char fn[1000];
  int i,j;
  FILE *f;
  sprintf(fn,"%s-%.5f-factors.dat",infilename,MAX_TRIANGLE);
  f=fopen(fn,"r");
  if (f==NULL) return 0;
  for (i=0;i<TriangleCount;i++) for (j=0;j<TriangleCount;j++) fscanf(f,"%lf ",&FF[i][j]);
  fclose(f);
  return 1;
}

void SaveFormFactors(void){
  char fn[1000];
  int i,j;
  FILE *f;
  sprintf(fn,"%s-%.5f-factors.dat",infilename,MAX_TRIANGLE);
  f=fopen(fn,"w");
  for (i=0;i<TriangleCount;i++) for (j=0;j<TriangleCount;j++) fprintf(f,"%g ",FF[i][j]);
  fclose(f);
}

void CalculateFormFactors(void){
  int T1,T2;
  int i,ray;
  double u1,v1,u2,v2;
  tPoint P1,P2;
  tVector V1,N1,N2;
  double nearest,where; int co;
  double cos1,cos2,dist;

  if (LoadFormFactors()) return;
  
  for (T1=0;T1<TriangleCount;T1++) {
    message("... for triangle %d\n",T1);
  for (T2=0;T2<TriangleCount;T2++) if (T1==T2) FF[T1][T2]=0.0; else {
    // pocitame form faktor pre dva rozne trojuholniky T1, T2
    // pouzijeme monte-carlo metodu
    // castneme niekolko lucov z jednej zaplaty na druhu, tie, co sa trafia, zratame
    FF[T1][T2]=0.0;
    if (!DO_MONTE_CARLO) MONTE_CARLO_RAYS=1;
    for (ray=0;ray<MONTE_CARLO_RAYS;ray++) {
      if (DO_MONTE_CARLO) {
        u1=((double)rand())/RAND_MAX; v1=((double)rand())/RAND_MAX; if (u1+v1>1.0) { u1=1.0-u1; v1=1.0-v1; }
        u2=((double)rand())/RAND_MAX; v2=((double)rand())/RAND_MAX; if (u2+v2>1.0) { u2=1.0-u2; v1=1.0-v2; }
      } else {
        u1=v1=u2=v2=1.0/3.0;
      }
      for (i=0;i<3;i++) P1[i]=P2[i]=0.0;
      for (i=0;i<3;i++) P1[i]+=TriangleList[T1].vertex[0][i]*u1;
      for (i=0;i<3;i++) P1[i]+=TriangleList[T1].vertex[1][i]*v1;
      for (i=0;i<3;i++) P1[i]+=TriangleList[T1].vertex[2][i]*(1-u1-v1);
      for (i=0;i<3;i++) P2[i]+=TriangleList[T2].vertex[0][i]*u2;
      for (i=0;i<3;i++) P2[i]+=TriangleList[T2].vertex[1][i]*v2;
      for (i=0;i<3;i++) P2[i]+=TriangleList[T2].vertex[2][i]*(1-u2-v2);
      for (i=0;i<3;i++) V1[i]=P2[i]-P1[i];
      nearest=INFTY;
      for (i=0;i<TriangleCount;i++) { 
        IntersectTriangle(i,P1,V1,&where);  
        if ((where>EPS) && (where<nearest)) { nearest=where; co=i; } 
      }  
      if (co!=T2) continue; // luc narazil po ceste na nieco ine
      InterpolateTriangleNormal(T1,P1,&N1);
      InterpolateTriangleNormal(T2,P2,&N2);
      Normalize(&V1);
      cos1=DotProduct(N1,V1);
      for (i=0;i<3;i++) V1[i]=-V1[i]; 
      cos2=DotProduct(N2,V1);
      if (cos1<0.0) cos1=0.0; if (cos2<0.0) cos2=0.0; // sme z blbej strany plochy
      dist=0.0; for (i=0;i<3;i++) dist+=sqr(P1[i]-P2[i]); dist+=EPS; // dist -- stvorec vzdial.
      FF[T1][T2]+=cos1*cos2/(dist*M_PI);
    }
    FF[T1][T2]/=MONTE_CARLO_RAYS;
//    for (i=0;i<3;i++) N1[i]=TriangleList[T2].vertex[1][i]-TriangleList[T2].vertex[0][i];
//    for (i=0;i<3;i++) N2[i]=TriangleList[T2].vertex[2][i]-TriangleList[T2].vertex[0][i];
//    CrossProduct(N1,N2,&V1);
    FF[T1][T2]*=Tarea[T1];
//    for (i=0;i<3;i++) N1[i]=TriangleList[T1].vertex[1][i]-TriangleList[T1].vertex[0][i];
//    for (i=0;i<3;i++) N2[i]=TriangleList[T1].vertex[2][i]-TriangleList[T1].vertex[0][i];
//    CrossProduct(N1,N2,&V1);
//    FF[T1][T2]*=Length(V1)/2.0;
//    FF[T1][T2]*=Length(V1)/2.0;
    message("FF[%d][%d]=%.5f\n",T1,T2,FF[T1][T2]);
  }}
  SaveFormFactors();
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   RADIOSITY                                               //
///////////////////////////////////////////////////////////////////////////////////////////////

void RadiosityStep(void){
  int best,i,j,k;
  double sbest,tu;
  // prejdeme vsetky plochy, najdeme tu, co chce najviac vyziarit
  best=0; sbest=TriangleList[0].emission[0]+TriangleList[0].emission[1]+TriangleList[0].emission[2];
  for (i=1;i<TriangleCount;i++) {
    tu=TriangleList[i].emission[0]+TriangleList[i].emission[1]+TriangleList[i].emission[2];
    if (tu>sbest) { sbest=tu; best=i; }
  }
  i=best;
  for (j=0;j<TriangleCount;j++) {
    for (k=0;k<3;k++) {
      TriangleList[j].emission[k]+=TriangleList[j].diffuseColor[k]*TriangleList[i].emission[k]*FF[i][j];
      TriangleList[j].excidentLight[k]+=TriangleList[j].diffuseColor[k]*TriangleList[i].emission[k]*FF[i][j];
//      message("Transferred energy: %g\n",TriangleList[j].diffuseColor[k]*TriangleList[i].emission[k]*FF[j][i]);
    }
  }
  for (k=0;k<3;k++) TriangleList[i].emission[k]=0.0;
}

int TheSame(tPoint P1, tPoint P2){
  double dist;
  dist=sqr(P1[0]-P2[0])+sqr(P1[1]-P2[1])+sqr(P1[2]-P2[2]);
  if (dist<0.001) return 1;
  return 0;
}

void VertexColors1(void){
  int T1,T2,i,j,k,l;
  int pocet[3],bol1[3],bol2[3];

  for (T1=0;T1<TriangleCount;T1++){
    // poratame farby vo vrcholoch T1
    for (k=0;k<3;k++) 
      TriangleList[T1].vertexColor[0][k]=TriangleList[T1].vertexColor[1][k]=
      TriangleList[T1].vertexColor[2][k]=TriangleList[T1].excidentLight[k];
    pocet[0]=pocet[1]=pocet[2]=1;
    if (!SMOOTH) continue;
      
    for (T2=0;T2<TriangleCount;T2++) if (TriangleList[T1].id==TriangleList[T2].id) {
      // ak maju spolocnu stranu a rovnake id, prispeje nam
      bol1[0]=bol1[1]=bol1[2]=0; bol2[0]=bol2[1]=bol2[2]=0;
      for (i=0;i<3;i++) for (j=0;j<3;j++) 
        if (TheSame(TriangleList[T1].vertex[i],TriangleList[T2].vertex[j])) { bol1[i]=1; bol2[j]=1; }
      if ((bol1[0]+bol1[1]+bol1[2])>=1) {
          // priratame
          for (k=0;k<3;k++) {
            pocet[k]+=bol1[k]; 
            for (l=0;l<3;l++) TriangleList[T1].vertexColor[k][l]+=bol1[k]*TriangleList[T2].excidentLight[l];
          }
      }
    }
    // a spriemerujeme
    for (i=0;i<3;i++) for (j=0;j<3;j++) TriangleList[T1].vertexColor[i][j]/=pocet[i];
  }
}

void VertexColors2(void){
  int T1,T2,i,j,k,l;
  int pocet[3],bol1[3],bol2[3];
/*  tPoint stred,P1,P2;
  tVector V1,V2;*/

  for (T1=0;T1<TriangleCount;T1++){
    // poratame farby vo vrcholoch T1
    for (k=0;k<3;k++) 
      TriangleList[T1].vertexColor[0][k]=TriangleList[T1].vertexColor[1][k]=
      TriangleList[T1].vertexColor[2][k]=TriangleList[T1].excidentLight[k];
    pocet[0]=pocet[1]=pocet[2]=1;
      
    for (T2=0;T2<TriangleCount;T2++) if (TriangleList[T1].id==TriangleList[T2].id) {
      // ak maju spolocnu stranu a su podobne natocene, prispeje nam
      bol1[0]=bol1[1]=bol1[2]=0; bol2[0]=bol2[1]=bol2[2]=0;
      for (i=0;i<3;i++) for (j=0;j<3;j++) 
        if (TheSame(TriangleList[T1].vertex[i],TriangleList[T2].vertex[j])) { bol1[i]=1; bol2[j]=1; }
      if ((bol1[0]+bol1[1]+bol1[2])==2) {
/*        // este to natocenie
        for (k=0;k<3;k++) { 
          stred[k]=(bol1[0]*TriangleList[T1].vertex[0][k]+
            bol1[1]*TriangleList[T1].vertex[1][k]+bol1[2]*TriangleList[T1].vertex[2][k])*0.5;
          P1[k]=(1-bol1[0])*TriangleList[T1].vertex[0][k]+
            (1-bol1[1])*TriangleList[T1].vertex[1][k]+(1-bol1[2])*TriangleList[T1].vertex[2][k];
          P2[k]=(1-bol2[0])*TriangleList[T2].vertex[0][k]+
            (1-bol2[1])*TriangleList[T2].vertex[1][k]+(1-bol2[2])*TriangleList[T2].vertex[2][k];
          V1[k]=P1[k]-stred[k];
          V2[k]=P2[k]-stred[k];
        }
        Normalize(&V1); Normalize(&V2);
        if (DotProduct(V1,V2)<-0.5) {*/
          // priratame
          for (k=0;k<3;k++) {
            pocet[k]+=bol1[k]; 
            for (l=0;l<3;l++) TriangleList[T1].vertexColor[k][l]+=bol1[k]*TriangleList[T2].excidentLight[l];
          }
/*        }*/
      }
    }
    // a spriemerujeme
    for (i=0;i<3;i++) for (j=0;j<3;j++) TriangleList[T1].vertexColor[i][j]/=pocet[i];
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////
//                                   MAIN                                                    //
///////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv){
  int i,pom;
  double dpom;

  infilename[0]=0;
  for (i=1;i<argc;i++){
    if (!strcmp(argv[i],"-v")) { VERBOSE=1; continue; }
    if (!strcmp(argv[i],"-d")) { DEBUG=1; continue; }
    if (!strcmp(argv[i],"-nomc")) { DO_MONTE_CARLO=0; continue; }
    if (sscanf(argv[i],"-smooth%d",&pom)) { SMOOTH=pom; continue; }
    if (sscanf(argv[i],"-x%d",&pom)) { XSIZE=pom; continue; }
    if (sscanf(argv[i],"-y%d",&pom)) { YSIZE=pom; continue; }
    if (sscanf(argv[i],"-i%d",&pom)) { POCET_ITERACII=pom; continue; }
    if (sscanf(argv[i],"-t%lf",&dpom)) { MAX_TRIANGLE=dpom; continue; }
    if (sscanf(argv[i],"-mcr%d",&pom)) { MONTE_CARLO_RAYS=pom; continue; }
    if (!strlen(infilename)) strcpy(infilename,argv[i]);
  }

  message("Loading scene from file '%s'.\n",infilename);
  LoadScene();
  message("%d triangles loaded.\n",TriangleCount);
  message("Calculating form factors.\n");
  CalculateFormFactors();
  for (iteracia=0;iteracia<POCET_ITERACII;iteracia++) {
    RadiosityStep();
/*    message("---\n");
    for (i=0;i<TriangleCount;i++) message("%3d: excident [%.4f %.4f %.4f] to emit [%.4f %.4f %.4f]\n",
        i,
        TriangleList[i].excidentLight[0],
        TriangleList[i].excidentLight[1],
        TriangleList[i].excidentLight[2],
        TriangleList[i].emission[0],
        TriangleList[i].emission[1],
        TriangleList[i].emission[2]);*/
    message("Radiosity iteration: %d\r",iteracia+1);
  }
  message("\n");
  if (SMOOTH>1) VertexColors2(); else VertexColors1();
  RayTrace();

  return 0;
}
